/*
  led.c
  Matlib 2010-02-04
*/

#include "rs232.c"

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <signal.h>

#include <arguments.h>

#include "led.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#include <semaphore.h>

  
#include "net.c"



struct DevData *dlist;

struct DevData *devdata_new (char *name, void (*led_set) (struct DevData *),
   void (*led_clr) (struct DevData *))
{
  struct DevData *dd, **i;
  dd = malloc (sizeof (struct DevData));
  CHECK_ERRORS (dd == NULL, "malloc");
  dd -> next = NULL;
  dd -> snext = NULL;
  sem_init (&(dd -> slock), 0, 1);
  dd -> name = name;
  
  dd -> fd = -1;
  dd -> arg = 0;
  
  sem_init (&(dd -> sem), 0, 1);
  dd -> active_blink = NULL;
  dd -> led_set = led_set;
  dd -> led_clr = led_clr;
  
  for (i = &dlist; *i != NULL; i = &((*i) -> next));
  *i = dd;
  return dd;
}

// sub-list handling
uint8_t slist_add (struct DevData **slist, int index)
{
  struct DevData *dd;
  dd = dlist;
  while (index > 0)
  {
    if (dd == NULL)
      return 2;
    dd = dd -> next;
    index --;
  }
  if (sem_trywait (&dd -> slock) != 0)
    return 1;
  dd -> snext = NULL;
  while (*slist != NULL)
    slist = &(*slist) -> snext;
  *slist = dd;
  return 0;
}

void slist_free (struct DevData **slist)
{
  struct DevData *snext;
  while (*slist != NULL)
  {
    snext = (*slist) -> next;
    (*slist) -> snext = NULL;
    sem_post (&(*slist) -> slock);
    *slist = snext;
  }
  *slist = NULL;
}







/* led_blink (<DevData>)
   Lights up one LED for the specified time (dd -> delay). 
   dd -> fill (0-100) selects PWM fill ratio. Current pattern 
   index is dd -> pattern_pos */
void led_blink (struct DevData *dd)
{
  struct timespec ts1, ts2;
  int t, e;
  switch (dd -> fill [dd -> pattern_pos])
  {
    case 0:
      ts1.tv_sec = dd -> delay [dd -> pattern_pos] / 100;
      ts1.tv_nsec = 1e7 * (dd -> delay [dd -> pattern_pos] % 100);
      nanosleep (&ts1, NULL);
      break;
    case 100:
      ts1.tv_sec = dd -> delay [dd -> pattern_pos] / 100;
      ts1.tv_nsec = 1e7 * (dd -> delay [dd -> pattern_pos] % 100);
      dd -> led_set (dd);
      nanosleep (&ts1, NULL);
      dd -> led_clr (dd);
      break;
    default:
      ts1.tv_sec = 0;
      ts1.tv_nsec = 1e5 * dd -> fill [dd -> pattern_pos];
      ts2.tv_sec = 0;
      ts2.tv_nsec = 1e5 * (100 - dd -> fill [dd -> pattern_pos]);
      e = 0;
      for (t = 0; (e == 0) && (t < dd -> delay [dd -> pattern_pos]); t ++)
      {
        dd -> led_set (dd);
        e += nanosleep (&ts1, NULL);
        dd -> led_clr (dd);
        e += nanosleep (&ts2, NULL);
      }
  }
}



#include "util.c"


/* Threads */

#define DD ((struct DevData *) dd)


/* blink_once
   Plays blink sequence once */
void *blink_once (void *dd)
{
  pthread_cleanup_push (sem_clean, &(DD -> sem));
  for (DD -> pattern_pos = 0; DD -> pattern_pos < DD -> pattern_length; 
     DD -> pattern_pos ++)
    led_blink (DD);
  pthread_cleanup_pop (1);
  return NULL;
}

/* blink_loop
   Plays blink sequence continuously. */
void *blink_loop (void *dd)
{
  signal (SIGUSR1, sig_ignore);
  pthread_cleanup_push (sem_clean, &(DD -> sem));
  DD -> pattern_pos = 0;
loop:
  while (DD -> pattern_pos < DD -> pattern_length)
  {
    led_blink (DD);
    DD -> pattern_pos ++;
  }
  DD -> pattern_pos = DD -> loop;
  goto loop;
  pthread_cleanup_pop (1);
}

/* blink_dim
   Lights the LED up with PWM fill percentage stored in dd -> loop. */
void *blink_dim (void *dd)
{
  struct timespec ts1, ts2;
  pthread_cleanup_push (sem_clean, &(DD -> sem));
  ts1.tv_sec = 0;
  ts2.tv_sec = 0;
loop:
  ts1.tv_nsec = 1e5 * DD -> loop;
  ts2.tv_nsec = 1e5 * (100 - DD -> loop);
  DD -> led_set (DD);
  nanosleep (&ts1, NULL);
  DD -> led_clr (DD);
  nanosleep (&ts2, NULL);
  goto loop;
  pthread_cleanup_pop (1);
}

#undef DD


/* kill_pending
   Kill working blink_* thread */
void kill_pending (struct DevData *dd)
{
  if (sem_trywait (&dd -> sem) != 0)
  {
    pthread_cancel (dd -> blink_tid);
    sem_wait (&dd -> sem);
  }
}

/* blink_start
   Start a new working thread */
void blink_start (struct DevData *dd, void * (*start_routine)(void *))
{
  pthread_create (&dd -> blink_tid, &thread_attr, start_routine, dd);
  dd -> active_blink = start_routine;
}




/* Connection handling */

void serve_connection (int *fd, struct DevData **slist)
{
  uint8_t cmd;
  uint8_t fill [PATTERN_MAX];
  uint16_t delay [PATTERN_MAX];
  uint32_t id;
  int pl, n;
  struct DevData *i;
  
cloop:
  read_data (fd, &cmd, 1);
  if (cmd == 'l')
  {
    slist_free (slist);
    do
    {
      read_data (fd, &id, 4);
      if (*fd == -1)
        return;
      cmd = slist_add (slist, id);
      write (*fd, &cmd, 1);
      if (cmd != 0)
        return;
      read_data (fd, &cmd, 1);
    }
    while (cmd == 'l');
  }
  if (*slist == NULL)
    return;

  pl = 0;
  while (*fd != -1)
    switch (cmd)
    {
      case 'B':
        if (pl == PATTERN_MAX)
          return;
        read_data (fd, fill + pl, 1);
        read_data (fd, delay + pl, 2);
        if ((fill [pl] > 100) || (delay [pl] < 50))
          return;
        pl ++;
        read_data (fd, &cmd, 1);
        break;
      case 'S':
        read_data (fd, &cmd, 1);
        if (cmd > 100)
          return;
        if (cmd % 100 == 0)
        {
          if (cmd == 0)
            for (i = *slist; i != NULL; i = i -> snext)
            {
              kill_pending (i);
              sem_post (&i -> sem);
              i -> led_clr (i);
            }
          else
            for (i = *slist; i != NULL; i = i -> snext)
            {
              kill_pending (i);
              sem_post (&i -> sem);
              i -> led_set (i);
            }
        }
        else
          for (i = *slist; i != NULL; i = i -> snext)
          {
            i -> loop = cmd;
            if (i -> active_blink != blink_dim)
            {
              kill_pending (i);
              blink_start (i, blink_dim);
            }
          }
        goto cloop;
      case 'L':
        read_data (fd, &cmd, 1);
        if ((pl == 0) || (cmd >= pl))
          return;
        for (i = *slist; i != NULL; i = i -> snext)
        {
          for (n = 0; n < pl; n ++)
          {
            i -> fill [n] = fill [n];
            i -> delay [n] = delay [n] / 10;
          }
          i -> loop = cmd;
          i -> pattern_length = pl;
          i -> pattern_pos = pl;
          if (i -> active_blink == blink_loop)
            pthread_kill (i -> blink_tid, SIGUSR1);
          else
          {
            kill_pending (i);
            blink_start (i, blink_loop);
          }
        }
        goto cloop;
      case 'G':
        if (pl == 0)
          return;
        for (i = *slist; i != NULL; i = i -> snext)
        {
          kill_pending (i);
          for (n = 0; n < pl; n ++)
          {
            i -> fill [n] = fill [n];
            i -> delay [n] = delay [n] / 10;
          }
          i -> pattern_length = pl;
          blink_start (i, blink_once);
        }
        goto cloop;
      default:
        return;
    }
}

void *connection_thread (void *p)
{
  struct DevData *slist;
  int fd;
  slist = NULL;
  fd = (long int) p;
  serve_connection (&fd, &slist);
  slist_free (&slist);
  close (fd);
  return NULL;
}



void daemon_mode ()
{
  pid_t pid;

  pid = fork ();
  CHECK_ERRORS (pid == -1, "fork");
  if (pid != 0)
    exit (EXIT_SUCCESS);
  fclose (stdin);
}

void blink_all ()
{
  struct DevData *dd;
  for (dd = dlist; dd != NULL; dd = dd -> next)
    dd -> led_set (dd);
  sleep (1);
  for (dd = dlist; dd != NULL; dd = dd -> next)
    dd -> led_clr (dd);  
}


struct Argument arguments [] =
{
  {'f', "foreground", argument_bool,   ARGUMENT_F_ONCE, NULL},
  {'n', "no-blink",   argument_bool,   ARGUMENT_F_ONCE, NULL},
  {'p', "port",       argument_string, ARGUMENT_F_PAR | ARGUMENT_F_ONCE, "335"},
  {}
};
#define ARG_FGND (arguments [0].value != NULL)
#define ARG_NBNK ((long int) arguments [1].value)
#define ARG_PORT ((char *) arguments [2].value)


int main (int argc, char **argv)
{
  int i, fd;
  pthread_t tid;
  
  block_signals (!ARG_FGND);
  pthread_attr_init (&thread_attr);
  pthread_attr_setdetachstate (&thread_attr, PTHREAD_CREATE_DETACHED);
  dlist = NULL;
  i = arguments_parse ("led", "0", "device ...", 
       "LED driver", argc, argv, arguments);
  if (i == argc)
    arguments_error_freeform ("Device node name is expected");
  do
  {
    rs232_init (argv [i]);
    i ++;
  }
  while (i < argc);
  
  i = make_socket (ARG_PORT);
  if (!ARG_NBNK)
    blink_all ();
  
  if (!ARG_FGND)
    daemon_mode ();
loop:
  fd = accept (i, NULL, 0);
  CHECK_ERRORS (fd == -1, "accept");
  pthread_create (&tid, &thread_attr, connection_thread, 
     (void *) ((long int) fd));
  goto loop;
}

